import json
from rich.console import Console
from rich.table import Table
from rich import print_json
from rich import print as rprint
import typer
from src.cpr_sdk.search_adaptors import VespaSearchAdapter
from src.cpr_sdk.models.search import SearchParameters
from tests.conftest import VESPA_TEST_SEARCH_URL


def main(
    instance_url: str = VESPA_TEST_SEARCH_URL,
    exact_match: bool = False,
    limit: int = 10,
):
    """Run a search query with different rank profiles."""
    console = Console()
    search_adapter = VespaSearchAdapter(instance_url)

    while True:
        query = input("Enter your search query (or 'q' to quit): ")
        if query.lower() == "q":
            break

        search_parameters = SearchParameters(
            query_string=query, exact_match=exact_match, limit=limit
        )
        search_response = search_adapter.search(search_parameters)

        for family in search_response.families:
            family_data = family.hits[0].model_dump()
            console.rule(
                title=f"{family_data['family_name']} ({family_data['family_geography']} ,{family_data['family_import_id']})"
            )
            print_json(
                json.dumps(
                    {
                        k: v
                        for k, v in family_data.items()
                        if not k.startswith("text_block") and "metadata" not in k
                    },
                    default=str,
                )
            )

            # There's some typing weirdness going on here:
            # hasattr(family.hits[0], 'text_blocks') can be False, but
            # family.hits[0].text_block exists
            try:
                rprint("Text blocks:")
                table = Table(title="Hits Table")

                # Add columns to the table
                table.add_column(
                    "Text Block ID", justify="right", style="cyan", no_wrap=True
                )
                table.add_column("Text Block", style="magenta")

                # Add rows to the table
                for hit in family.hits:
                    try:
                        table.add_row(f"{hit.text_block_id}", f"{hit.text_block}")  # type: ignore
                    except Exception:
                        pass

                # Print the table
                console.print(table)

                if family == search_response.families[-1]:
                    print("No more families to show.")
                    break
            except AttributeError:
                print("No text blocks found.")

            input("Press any key to show next family")


if __name__ == "__main__":
    typer.run(main)



import re
from datetime import datetime
from typing import List, Literal, Optional, Sequence

from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_validator,
    model_validator,
)

# Value Lookup Tables
sort_orders = {
    "asc": "+",
    "desc": "-",
    "ascending": "+",
    "descending": "-",
}

sort_fields = {
    "date": "family_publication_ts",
    "title": "family_name",
    "name": "family_name",
}

filter_fields = {
    "geography": "family_geography",
    "geographies": "family_geographies",
    "category": "family_category",
    "language": "document_languages",
    "source": "family_source",
}

_ID_ELEMENT = r"[a-zA-Z0-9]+([-_]?[a-zA-Z0-9]+)*"
ID_PATTERN = re.compile(rf"{_ID_ELEMENT}\.{_ID_ELEMENT}\.{_ID_ELEMENT}\.{_ID_ELEMENT}")


class MetadataFilter(BaseModel):
    """A filter for metadata fields"""

    name: str
    value: str


class ConceptFilter(BaseModel):
    """A filter for concept fields"""

    name: Literal["name", "id", "model", "timestamp", "parent_concept_ids_flat"]
    value: str

    @model_validator(mode="after")
    def validate_parent_concept_ids_flat(self) -> "ConceptFilter":
        """
        Validate parent_concept_ids_flat field.

        In the schema we comma separate values in the parent_concept_ids_flat field.
        This means we must ensure that the last character is a comma to avoid the
        situation below:

        E.g. querying parent_concept_ids_flat on "Q1" should only return "Q1" but you
        will also return "Q12, Q123" which is invalid.

        To get around this we query on "Q1," instead using the comma suffix to separate
        values.
        """
        if self.name == "parent_concept_ids_flat" and self.value[-1] != ",":
            self.value = self.value + ","
        return self


class Concept(BaseModel):
    """
    A concept extracted from a passage of text.

    This refers to a span of text within passage that holds a concept.
    E.g. "Adaptation strategy" is a concept within a passage starting at index 0 and
    ending at index 17, classified by model "environment_model_1" on the 12th Jan at
    12:00.
    """

    id: str
    name: str
    parent_concepts: List[dict[str, str]]
    parent_concept_ids_flat: str
    model: str
    end: int
    start: int
    timestamp: datetime

    model_config = ConfigDict(
        use_enum_values=True,
        json_encoders={datetime: lambda dt: dt.isoformat()},
    )

    @model_validator(mode="after")
    def validate_parent_concept_ids_flat(self) -> "Concept":
        """
        Validate parent_concept_ids_flat field.

        This field should hold the same ids as concepts in the parent_concepts field.
        """
        parent_concept_ids_flattened = ",".join(
            [parent_concept["id"] for parent_concept in self.parent_concepts]
        )

        if not (
            self.parent_concept_ids_flat == parent_concept_ids_flattened
            or self.parent_concept_ids_flat == parent_concept_ids_flattened + ","
        ):
            raise ValueError(
                "parent_concept_ids_flat must be a comma separated list of parent "
                "concept ids held in the parent concepts field. "
                f"Received parent_concept_ids_flat: {self.parent_concept_ids_flat}\n"
                "Received ids in the parent_concept objects: "
                f"{parent_concept_ids_flattened}"
            )
        return self


class Filters(BaseModel):
    """Filterable fields in a search request"""

    family_geography: Sequence[str] = []
    family_geographies: Sequence[str] = []
    family_category: Sequence[str] = []
    document_languages: Sequence[str] = []
    family_source: Sequence[str] = []

    model_config: ConfigDict = {
        "extra": "forbid",
    }

    @field_validator(
        "family_geographies",
        "family_geography",
        "family_category",
        "document_languages",
        "family_source",
    )
    def sanitise_filter_inputs(cls, field):
        """Remove problematic characters from filter values"""
        clean_values = []
        for keyword in field:
            keyword = keyword.replace('"', "")
            keyword = keyword.replace("\\", " ")
            keyword = " ".join(keyword.split())
            clean_values.append(keyword)
        return clean_values


class SearchParameters(BaseModel):
    """Parameters for a search request"""

    query_string: Optional[str] = ""
    """
    A string representation of the search to be performed.
    For example: 'Adaptation strategy'"
    """

    exact_match: bool = False
    """
    Indicate if the `query_string` should be treated as an exact match when
    the search is performed.
    """

    all_results: bool = False
    """
    Return all results rather than searching or ranking

    Filters can still be applied
    """

    documents_only: bool = False
    """Ignores passages in search when true."""

    limit: int = Field(ge=0, default=100, le=500)
    """
    Refers to the maximum number of results to return from the "
    query result.
    """

    max_hits_per_family: int = Field(
        validation_alias=AliasChoices("max_passages_per_doc", "max_hits_per_family"),
        default=10,
        ge=0,
        le=500,
    )
    """
    The maximum number of matched passages to be returned for a "
    single document.
    """

    family_ids: Optional[Sequence[str]] = None
    """Optionally limit a search to a specific set of family ids."""

    document_ids: Optional[Sequence[str]] = None
    """Optionally limit a search to a specific set of document ids."""

    filters: Optional[Filters] = None
    """Filter results to matching filter items."""

    year_range: Optional[tuple[Optional[int], Optional[int]]] = None
    """
    The years to search between. Containing exactly two values,
    which can be null or an integer representing the years to
    search between. These are inclusive and can be null. Example:
    [null, 2010] will return all documents return in or before 2010.
    """

    sort_by: Optional[str] = Field(
        validation_alias=AliasChoices("sort_field", "sort_by"), default=None
    )
    """The field to sort by can be chosen from `date` or `title`."""

    sort_order: str = "descending"
    """
    The order of the results according to the `sort_field`, can be chosen from
    ascending (use “asc”) or descending (use “desc”).
    """

    continuation_tokens: Optional[Sequence[str]] = None
    """
    Use to return the next page of results from a specific search, the next token
    can be found on the response object. It's also possible to get the next page
    of passages by including the family level continuation token first in the
    array followed by the passage level one.
    """

    corpus_type_names: Optional[Sequence[str]] = None
    """
    The name of the corpus that a document belongs to.
    """

    corpus_import_ids: Optional[Sequence[str]] = None
    """
    The import id of the corpus that a document belongs to.
    """

    metadata: Optional[Sequence[MetadataFilter]] = None
    """
    A field and item mapping to search in the metadata field of the documents.

    E.g. [{"name": "family.sector", "value": "Price"}]
    """

    concept_filters: Optional[Sequence[ConceptFilter]] = None
    """
    A field and item mapping to search in the concepts field of the document passages.
    """

    @model_validator(mode="after")
    def validate(self):
        """Validate against mutually exclusive fields"""
        if self.exact_match and self.all_results:
            raise ValueError("`exact_match` and `all_results` are mutually exclusive")
        if self.documents_only and not self.all_results:
            raise ValueError(
                "`documents_only` requires `all_results`, other queries are not supported"
            )
        return self

    @model_validator(mode="after")
    def concept_filters_not_set_if_documents_only(self) -> "SearchParameters":
        """Ensure concept_filters are not set if browse mode (documents_only) is set."""
        if self.concept_filters is not None and self.documents_only is True:
            raise ValueError(
                "Cannot set concept_filters when only searching documents. This is as concept_filters are only applicable to passages."
            )
        return self

    @field_validator("continuation_tokens")
    def continuation_tokens_must_be_upper_strings(cls, continuation_tokens):
        """Validate continuation_tokens match the expected format"""
        if not continuation_tokens:
            return continuation_tokens

        for token in continuation_tokens:
            if token == "":
                continue
            if not token.isalpha():
                raise ValueError(f"Expected continuation tokens to be letters: {token}")
            if not token.isupper():
                raise ValueError(
                    f"Expected continuation tokens to be uppercase: {token}"
                )
        return continuation_tokens

    @model_validator(mode="after")
    def query_string_must_not_be_empty(self):
        """Validate that the query string is not empty."""
        if not self.query_string:
            self.all_results = True
        return self

    @field_validator("family_ids", "document_ids")
    def ids_must_fit_pattern(cls, ids):
        """
        Validate that the family and document ids are ids.

        Example ids:
            CCLW.document.i00000004.n0000
            CCLW.family.i00000003.n0000
            CCLW.executive.10014.4470
            CCLW.family.10014.0
        """
        if ids:
            for _id in ids:
                if not re.fullmatch(ID_PATTERN, _id):
                    raise ValueError(f"id seems invalid: {_id}")
        return ids

    @field_validator("year_range")
    def year_range_must_be_valid(cls, year_range):
        """Validate that the year range is valid."""
        if year_range is not None:
            if year_range[0] is not None and year_range[1] is not None:
                if year_range[0] > year_range[1]:
                    raise ValueError(
                        "The first supplied year must be less than or equal to the "
                        f"second supplied year. Received: {year_range}"
                    )
        return year_range

    @field_validator("sort_by")
    def sort_by_must_be_valid(cls, sort_by):
        """Validate that the sort field is valid."""
        if sort_by is not None:
            if sort_by not in sort_fields:
                raise ValueError(
                    f"Invalid sort field: {sort_by}. sort_by must be one of: "
                    f"{list(sort_fields.keys())}"
                )
        return sort_by

    @field_validator("sort_order")
    def sort_order_must_be_valid(cls, sort_order):
        """Validate that the sort order is valid."""
        if sort_order not in sort_orders:
            raise ValueError(
                f"Invalid sort order: {sort_order}. sort_order must be one of: "
                f"{sort_orders}"
            )
        return sort_order

    @computed_field
    def vespa_sort_by(self) -> Optional[str]:
        """Translates sort by into the format acceptable by vespa"""
        if self.sort_by:
            return sort_fields.get(self.sort_by)
        else:
            return None

    @computed_field
    def vespa_sort_order(self) -> Optional[str]:
        """Translates sort order into the format acceptable by vespa"""
        return sort_orders.get(self.sort_order)


class Hit(BaseModel):
    """Common model for all search result hits."""

    family_name: Optional[str] = None
    family_description: Optional[str] = None
    family_source: Optional[str] = None
    family_import_id: Optional[str] = None
    family_slug: Optional[str] = None
    family_category: Optional[str] = None
    family_publication_ts: Optional[datetime] = None
    family_geography: Optional[str] = None
    family_geographies: Optional[List[str]] = None
    document_import_id: Optional[str] = None
    document_slug: Optional[str] = None
    document_languages: Optional[List[str]] = None
    document_content_type: Optional[str] = None
    document_cdn_object: Optional[str] = None
    document_source_url: Optional[str] = None
    corpus_type_name: Optional[str] = None
    corpus_import_id: Optional[str] = None
    metadata: Optional[Sequence[dict[str, str]]] = None
    concepts: Optional[Sequence[Concept]] = None
    relevance: Optional[float] = None
    rank_features: Optional[dict[str, float]] = None

    @classmethod
    def from_vespa_response(cls, response_hit: dict) -> "Hit":
        """
        Create a Hit from a Vespa response hit.

        :param dict response_hit: part of a json response from Vespa
        :raises ValueError: if the response type is unknown
        :return Hit: an individual document or passage hit
        """
        # vespa structures its response differently depending on the api endpoint
        # for searches, the response should contain a sddocname field
        response_type = response_hit.get("fields", {}).get("sddocname")
        if response_type is None:
            # for get_by_id, the response should contain an id field
            response_type = response_hit["id"].split(":")[2]

        if response_type == "family_document":
            hit = Document.from_vespa_response(response_hit=response_hit)
        elif response_type == "document_passage":
            hit = Passage.from_vespa_response(response_hit=response_hit)
        else:
            raise ValueError(f"Unknown response type: {response_type}")
        return hit


class Document(Hit):
    """A document search result hit."""

    @classmethod
    def from_vespa_response(cls, response_hit: dict) -> "Document":
        """
        Create a Document from a Vespa response hit.

        :param dict response_hit: part of a json response from Vespa
        :return Document: a populated document
        """
        fields = response_hit["fields"]
        family_publication_ts = fields.get("family_publication_ts", None)
        family_publication_ts = (
            datetime.fromisoformat(family_publication_ts)
            if family_publication_ts
            else None
        )
        return cls(
            family_name=fields.get("family_name"),
            family_description=fields.get("family_description"),
            family_source=fields.get("family_source"),
            family_import_id=fields.get("family_import_id"),
            family_slug=fields.get("family_slug"),
            family_category=fields.get("family_category"),
            family_publication_ts=family_publication_ts,
            family_geography=fields.get("family_geography"),
            family_geographies=fields.get("family_geographies", []),
            document_import_id=fields.get("document_import_id"),
            document_slug=fields.get("document_slug"),
            document_languages=fields.get("document_languages", []),
            document_content_type=fields.get("document_content_type"),
            document_cdn_object=fields.get("document_cdn_object"),
            document_source_url=fields.get("document_source_url"),
            corpus_type_name=fields.get("corpus_type_name"),
            corpus_import_id=fields.get("corpus_import_id"),
            metadata=fields.get("metadata"),
            concepts=fields.get("concepts"),
            relevance=response_hit.get("relevance"),
            rank_features=fields.get("summaryfeatures"),
        )


class Passage(Hit):
    """A passage search result hit."""

    text_block: str
    text_block_id: str
    text_block_type: str
    text_block_page: Optional[int] = None
    text_block_coords: Optional[Sequence[tuple[float, float]]] = None

    @classmethod
    def from_vespa_response(cls, response_hit: dict) -> "Passage":
        """
        Create a Passage from a Vespa response hit.

        :param dict response_hit: part of a json response from Vespa
        :return Passage: a populated passage
        """
        fields = response_hit["fields"]
        family_publication_ts = fields.get("family_publication_ts")
        family_publication_ts = (
            datetime.fromisoformat(family_publication_ts)
            if family_publication_ts
            else None
        )

        return cls(
            family_name=fields.get("family_name"),
            family_description=fields.get("family_description"),
            family_source=fields.get("family_source"),
            family_import_id=fields.get("family_import_id"),
            family_slug=fields.get("family_slug"),
            family_category=fields.get("family_category"),
            family_publication_ts=family_publication_ts,
            family_geography=fields.get("family_geography"),
            family_geographies=fields.get("family_geographies", []),
            document_import_id=fields.get("document_import_id"),
            document_slug=fields.get("document_slug"),
            document_languages=fields.get("document_languages", []),
            document_content_type=fields.get("document_content_type"),
            document_cdn_object=fields.get("document_cdn_object"),
            document_source_url=fields.get("document_source_url"),
            corpus_type_name=fields.get("corpus_type_name"),
            corpus_import_id=fields.get("corpus_import_id"),
            text_block=fields["text_block"],
            text_block_id=fields["text_block_id"],
            text_block_type=fields["text_block_type"],
            text_block_page=fields.get("text_block_page"),
            text_block_coords=fields.get("text_block_coords"),
            metadata=fields.get("metadata"),
            concepts=fields.get("concepts"),
            relevance=response_hit.get("relevance"),
            rank_features=fields.get("summaryfeatures"),
        )


class Family(BaseModel):
    """A family containing relevant documents and passages."""

    id: str
    hits: Sequence[Hit]
    total_passage_hits: int = 0
    continuation_token: Optional[str] = None
    prev_continuation_token: Optional[str] = None
    relevance: Optional[float] = None


class SearchResponse(BaseModel):
    """Relevant results, and search response metadata"""

    total_hits: int
    total_family_hits: int = 0
    query_time_ms: Optional[int] = None
    total_time_ms: Optional[int] = None
    families: Sequence[Family]
    continuation_token: Optional[str] = None
    this_continuation_token: Optional[str] = None
    prev_continuation_token: Optional[str] = None



"""Adaptors for getting and storing data from CPR data sources."""

import logging
from abc import ABC, abstractmethod
from typing import List, Optional
from pathlib import Path
from tqdm.auto import tqdm

from cpr_sdk.parser_models import BaseParserOutput
from cpr_sdk.s3 import _get_s3_keys_with_prefix, _s3_object_read_text

_LOGGER = logging.getLogger(__name__)


class DataAdaptor(ABC):
    """Base class for data adaptors."""

    @abstractmethod
    def load_dataset(
        self, dataset_key: str, limit: Optional[int] = None
    ) -> List[BaseParserOutput]:
        """Load entire dataset from data source."""
        raise NotImplementedError

    @abstractmethod
    def get_by_id(
        self, dataset_key: str, document_id: str
    ) -> Optional[BaseParserOutput]:
        """Get a document by its id."""
        raise NotImplementedError


class S3DataAdaptor(DataAdaptor):
    """Adaptor for loading data from S3."""

    def load_dataset(
        self, dataset_key: str, limit: Optional[int] = None
    ) -> List[BaseParserOutput]:
        """
        Load entire dataset from S3.

        :param dataset_key: path to S3 directory. Should start with 's3://'
        :param limit: optionally limit number of documents loaded. Defaults to None
        :return List[BaseParserOutput]: list of parser outputs
        """
        if not dataset_key.startswith("s3://"):
            _LOGGER.warning(
                f"Dataset key {dataset_key} does not start with 's3://'. "
                "Assuming it is an S3 bucket."
            )
            dataset_key = f"s3://{dataset_key}"

        if dataset_key.endswith("/"):
            dataset_key = dataset_key[:-1]

        s3_objects = _get_s3_keys_with_prefix(dataset_key)

        if len(s3_objects) == 0:
            raise ValueError(f"No objects found at {dataset_key}.")

        parsed_files = []

        for filename in tqdm(s3_objects[:limit]):
            if filename.endswith(".json"):
                parsed_files.append(
                    BaseParserOutput.model_validate_json(
                        _s3_object_read_text(f"{dataset_key}/{filename.split('/')[-1]}")
                    )
                )

        return parsed_files

    def get_by_id(
        self, dataset_key: str, document_id: str
    ) -> Optional[BaseParserOutput]:
        """
        Get a document by its id.

        :param str dataset_key: S3 bucket
        :param str document_id: import ID
        :return Optional[BaseParserOutput]: None if no document was found with the ID
        """

        try:
            return BaseParserOutput.model_validate_json(
                _s3_object_read_text(f"s3://{dataset_key}/{document_id}.json")
            )
        except ValueError as e:
            if "does not exist" in str(e):
                return None
            else:
                raise e
        except Exception as e:
            raise e


class LocalDataAdaptor(DataAdaptor):
    """Adaptor for loading data from a local path."""

    def load_dataset(
        self, dataset_key: str, limit: Optional[int] = None
    ) -> List[BaseParserOutput]:
        """
        Load entire dataset from a local path.

        :param str dataset_key: path to local directory containing parser outputs/embeddings inputs
        :param limit: optionally limit number of documents loaded. Defaults to None
        :return List[BaseParserOutput]: list of parser outputs
        """

        folder_path = Path(dataset_key).resolve()

        if not folder_path.exists():
            raise ValueError(f"Path {folder_path} does not exist")

        if not folder_path.is_dir():
            raise ValueError(f"Path {folder_path} is not a directory")

        if len(list(folder_path.glob("*.json"))) == 0:
            raise ValueError(f"Path {folder_path} does not contain any json files")

        parsed_files = []

        files = list(folder_path.glob("*.json"))[:limit]
        num_batches = len(files) // 1000 + 1

        for batch_idx in range(num_batches):
            parsed_batch_files = self._load_files(
                files[batch_idx * 1000 : (batch_idx + 1) * 1000], batch_idx, num_batches
            )
            parsed_files += parsed_batch_files

        return parsed_files

    @staticmethod
    def _load_files(file_paths: list[Path], batch_idx: int, num_batches: int):
        """Loads the files within a batch with paths provided in file_paths."""
        parsed_files = []

        raw_files = (file.read_text() for file in file_paths)
        for raw_file_text in tqdm(
            raw_files,
            desc=f"Loading files from directory in batch {batch_idx + 1}/{num_batches}",
        ):
            parsed_files.append(BaseParserOutput.model_validate_json(raw_file_text))

        return parsed_files

    def get_by_id(
        self, dataset_key: str, document_id: str
    ) -> Optional[BaseParserOutput]:
        """
        Get a document by its id.

        :param str dataset_key: path to local directory containing parser outputs/embeddings inputs
        :param str document_id: import ID
        :return Optional[BaseParserOutput]: None if no document was found with the ID
        """
        # TODO: these "get_by_id"  methods are almost certainly not what we need
        # because we're unable to load more complex subclasses of BaseParserOutput

        folder_path = Path(dataset_key).resolve()

        if not folder_path.exists():
            raise ValueError(f"Path {folder_path} does not exist")

        file_path = folder_path / f"{document_id}.json"

        if not file_path.exists():
            return None

        return BaseParserOutput.model_validate_json(file_path.read_text())



from string import Template
from typing import Optional

from rich import print as rprint
from cpr_sdk.models.search import Filters, SearchParameters


class YQLBuilder:
    """Used to assemble yql queries"""

    yql_base = Template(
        """
        select * from sources $SOURCES
            where $WHERE_CLAUSE
        limit 0
        |
            $CONTINUATION
        all(
            group(family_import_id)
            output(count())
            max($LIMIT)
            $SORT
            each(
                output(count())
                max($MAX_HITS_PER_FAMILY)
                each(
                    output(
                        summary(search_summary)
                    )
                )
            )
        )
    """
    )

    def __init__(self, params: SearchParameters, sensitive: bool = False) -> None:
        self.params = params
        self.sensitive = sensitive

    def build_sources(self) -> str:
        """Creates the part of the query that determines which sources to search"""
        if self.params.documents_only:
            return "family_document"
        else:
            return "family_document, document_passage"

    def build_search_term(self) -> str:
        """Create the part of the query that matches a users search text"""
        if self.params.all_results:
            return "( true )"
        if self.params.exact_match:
            return """
                (
                    (family_name_not_stemmed contains({stem: false}@query_string)) or
                    (family_description_not_stemmed contains({stem: false}@query_string)) or
                    (text_block_not_stemmed contains ({stem: false}@query_string))
                )
            """
        elif self.sensitive:
            return """
                (
                    (userInput(@query_string)) 
                )
            """
        else:
            return """
                (
                    (userInput(@query_string)) 
                    or (
                        [{"targetNumHits": 1000}]
                        nearestNeighbor(family_description_embedding,query_embedding)
                    ) or (
                        [{"targetNumHits": 1000}]
                        nearestNeighbor(text_embedding,query_embedding)
                    )
                )
            """

    def build_metadata_filter(self) -> Optional[str]:
        """Create the part of the query that limits to specific metadata"""
        metadata_filters = []
        if self.params.metadata:
            [
                metadata_filters.append(
                    f"""
                    (
                        metadata contains sameElement(
                            name contains '{metadata.name}',
                            value contains '{metadata.value}'
                        )
                    )
                    """
                )
                for metadata in self.params.metadata
            ]
            return f"({' and '.join(metadata_filters)})"
        return None

    def build_concepts_filter(self) -> Optional[str]:
        """Create the part of the query that limits to specific concepts"""
        if self.params.concept_filters:
            concepts_query = []
            for concept in self.params.concept_filters:
                if concept.name == "parent_concept_ids_flat":
                    concepts_query.append(f"{concept.name} matches '{concept.value}'")
                else:
                    concepts_query.append(f"{concept.name} contains '{concept.value}'")
            return f"(concepts contains sameElement({', '.join(concepts_query)}))"
        return None

    def build_corpus_type_name_filter(self) -> Optional[str]:
        """Create the part of the query that limits to specific corpora"""
        if self.params.corpus_type_names:
            corpora = ", ".join([f"'{c}'" for c in self.params.corpus_type_names])
            return f"(corpus_type_name in({corpora}))"

    def build_corpus_import_ids_filter(self) -> Optional[str]:
        """Create the part of the query that limits to specific corpora import id"""
        if self.params.corpus_import_ids:
            corpora = ", ".join([f"'{c}'" for c in self.params.corpus_import_ids])
            return f"(corpus_import_id in({corpora}))"

    def build_family_filter(self) -> Optional[str]:
        """Create the part of the query that limits to specific families"""
        if self.params.family_ids:
            families = ", ".join([f"'{f}'" for f in self.params.family_ids])
            return f"(family_import_id in({families}))"
        return None

    def build_document_filter(self) -> Optional[str]:
        """Create the part of the query that limits to specific documents"""
        if self.params.document_ids:
            documents = ", ".join([f"'{d}'" for d in self.params.document_ids])
            return f"(document_import_id in({documents}))"
        return None

    def _inclusive_filters(self, filters: Filters, field_name: str) -> Optional[str]:
        values = getattr(filters, field_name)
        query_filters = []
        for value in values:
            query_filters.append(f'({field_name} contains "{value}")')
        if query_filters:
            return f"({' or '.join(query_filters)})"

    def build_year_start_filter(self) -> Optional[str]:
        """Create the part of the query that filters on a year range"""
        if self.params.year_range:
            start, _ = self.params.year_range
            if start:
                return f"(family_publication_year >= {start})"
        return None

    def build_year_end_filter(self) -> Optional[str]:
        """Create the part of the query that filters on a year range"""
        if self.params.year_range:
            _, end = self.params.year_range
            if end:
                return f"(family_publication_year <= {end})"
        return None

    def build_where_clause(self) -> str:
        """Create the part of the query that adds filters"""
        filters = []
        filters.append(self.build_search_term())
        filters.append(self.build_family_filter())
        filters.append(self.build_document_filter())
        filters.append(self.build_corpus_type_name_filter())
        filters.append(self.build_corpus_import_ids_filter())
        filters.append(self.build_metadata_filter())
        filters.append(self.build_concepts_filter())
        if f := self.params.filters:
            filters.append(self._inclusive_filters(f, "family_geographies"))
            filters.append(self._inclusive_filters(f, "family_geography"))
            filters.append(self._inclusive_filters(f, "family_category"))
            filters.append(self._inclusive_filters(f, "document_languages"))
            filters.append(self._inclusive_filters(f, "family_source"))
        filters.append(self.build_year_start_filter())
        filters.append(self.build_year_end_filter())
        return " and ".join([f for f in filters if f])  # Remove empty

    def build_continuation(self) -> str:
        """Create the part of the query that adds continuation tokens"""
        if self.params.continuation_tokens:
            continuations = ", ".join(f"'{c}'" for c in self.params.continuation_tokens)
            return f"{{ 'continuations': [{continuations}] }}"
        else:
            return ""

    def build_limit(self) -> int:
        """Create the part of the query limiting the number of families returned"""
        return self.params.limit

    def build_sort(self) -> str:
        """Creates the part of the query used for sorting by different fields"""
        sort_by = self.params.vespa_sort_by
        sort_order = self.params.vespa_sort_order

        if not sort_by or not sort_order:
            return ""
        return f"order({sort_order}max({sort_by}))"

    def build_max_hits_per_family(self) -> int:
        """Create the part of the query limiting passages within a family returned"""
        return self.params.max_hits_per_family

    def to_str(self) -> str:
        """Assemble the yql from parts using the template"""
        yql = self.yql_base.substitute(
            SOURCES=self.build_sources(),
            WHERE_CLAUSE=self.build_where_clause(),
            CONTINUATION=self.build_continuation(),
            LIMIT=self.build_limit(),
            SORT=self.build_sort(),
            MAX_HITS_PER_FAMILY=self.build_max_hits_per_family(),
        )
        return " ".join(yql.split())


if __name__ == "__main__":
    # YQL example
    params = SearchParameters(
        query_string="climate",
        exact_match=False,
        limit=10,
        max_hits_per_family=10,
        filters=Filters(
            **{"document_languages": ["value"], "family_source": ["value"]}
        ),
        year_range=(2000, 2020),
        continuation_tokens=None,
    )

    yql_new = YQLBuilder(
        params=params,
        sensitive=False,
    ).to_str()

    rprint(yql_new)



import logging
from pathlib import Path
from typing import Any, List, Optional

import yaml
from vespa.exceptions import VespaError
from vespa.io import VespaResponse

from cpr_sdk.exceptions import FetchError
from cpr_sdk.models.search import Family, Hit, SearchParameters, SearchResponse
from cpr_sdk.utils import dig, is_sensitive_query, load_sensitive_query_terms
from cpr_sdk.yql_builder import YQLBuilder

SENSITIVE_QUERY_TERMS = load_sensitive_query_terms()
_LOGGER = logging.getLogger(__name__)


def split_document_id(document_id: str) -> tuple[str, str, str]:
    """
    Split a document_id into its namespace, schema, and data_id components.

    IDs should be of the form: "id:namespace:schema::data_id"

    :param str document_id: a document id of the form "id:namespace:schema::data_id"
    :raises ValueError: if the document id is not of the expected form
    :return tuple[str, str, str]: the namespace, schema, and data_id components of the
        document_id
    """
    try:
        namespace_and_schema, data_id = document_id.split("::")
        _, namespace, schema = namespace_and_schema.split(":")
    except ValueError as e:
        raise ValueError(
            f'Failed to parse document id: "{document_id}". '
            'Document ids should be of the form: "id:namespace:schema::data_id"'
        ) from e
    return namespace, schema, data_id


def find_vespa_cert_paths() -> tuple[Optional[str], Optional[str]]:
    """
    Automatically find the certificate and key files for the vespa instance

    :raises FileNotFoundError: if the .vespa directory is not found in the home
        directory, or if the application name is not found in the config.yaml file
    :return tuple[Path, Path]: The paths to the certificate and key files, respectively
    """
    vespa_directory = Path.home() / ".vespa/"
    if not vespa_directory.exists():
        _LOGGER.warning(
            "Could not find .vespa directory in home directory when looking for certs."
        )
        return None, None

    vespa_config = vespa_directory / "config.yaml"
    if not vespa_config.exists():
        _LOGGER.warning(
            "Could not find config.yaml file in .vespa directory when looking for certs."
        )
        return None, None

    # read the config.yaml file to find the application name
    with open(vespa_directory / "config.yaml", "r", encoding="utf-8") as yaml_file:
        data = yaml.safe_load(yaml_file)
        if not data or "application" not in data:
            return None, None
        application_name = data["application"]

    cert_directory = vespa_directory / application_name

    cert_directory_certs = list(cert_directory.glob("*cert.pem"))
    cert_path = (
        str(list(cert_directory.glob("*cert.pem"))[0]) if cert_directory_certs else None
    )
    cert_directory_keys = list(cert_directory.glob("*key.pem"))
    key_path = (
        str(list(cert_directory.glob("*key.pem"))[0]) if cert_directory_keys else None
    )

    return cert_path, key_path


def build_vespa_request_body(parameters: SearchParameters) -> dict[str, str]:
    """Constructs the payload for a vespa query"""
    sensitive = is_sensitive_query(parameters.query_string, SENSITIVE_QUERY_TERMS)

    yql = YQLBuilder(params=parameters, sensitive=sensitive).to_str()
    vespa_request_body: dict[str, Any] = {
        "yql": yql,
        "timeout": "20",
        "ranking.softtimeout.factor": "0.7",
        "query_string": parameters.query_string,
    }

    if parameters.all_results:
        pass
    elif parameters.exact_match:
        vespa_request_body["ranking.profile"] = "exact_not_stemmed"
    elif sensitive:
        vespa_request_body["ranking.profile"] = "hybrid_no_closeness"
    else:
        vespa_request_body["ranking.profile"] = "hybrid"
        vespa_request_body[
            "input.query(query_embedding)"
        ] = "embed(msmarco-distilbert-dot-v5, @query_string)"

    return vespa_request_body


def parse_vespa_response(vespa_response: VespaResponse) -> SearchResponse:
    """
    Parse a vespa response into a SearchResponse object

    :param SearchParameters request: The user's original search request
    :param VespaResponse vespa_response: The response from the vespa instance
    :raises FetchError: if the vespa response status code is not 200, indicating an
        error in the query, or the vespa instance
    :return SearchResponse: a list of families, with response metadata
    """
    if vespa_response.status_code != 200:
        raise FetchError(
            f"Received status code {vespa_response.status_code}",
            status_code=vespa_response.status_code,
        )
    families: List[Family] = []
    root = vespa_response.json["root"]

    response_families = dig(root, "children", 0, "children", 0, "children", default=[])
    for family in response_families:
        total_passage_hits = dig(family, "fields", "count()")
        family_hits: List[Hit] = []
        passages_continuation = dig(family, "children", 0, "continuation", "next")
        prev_passages_continuation = dig(family, "children", 0, "continuation", "prev")
        family_relevance = family.get("relevance")
        for hit in dig(family, "children", 0, "children", default=[]):
            family_hits.append(Hit.from_vespa_response(response_hit=hit))
        families.append(
            Family(
                id=family["value"],
                hits=family_hits,
                total_passage_hits=total_passage_hits,
                continuation_token=passages_continuation,
                prev_continuation_token=prev_passages_continuation,
                relevance=family_relevance,
            )
        )

    next_family_continuation = dig(
        root, "children", 0, "children", 0, "continuation", "next"
    )
    prev_family_continuation = dig(
        root, "children", 0, "children", 0, "continuation", "prev"
    )
    this_family_continuation = dig(root, "children", 0, "continuation", "this")
    total_hits = dig(root, "fields", "totalCount", default=0)
    total_family_hits = dig(root, "children", 0, "fields", "count()", default=0)
    return SearchResponse(
        total_hits=total_hits,
        total_family_hits=total_family_hits,
        families=families,
        continuation_token=next_family_continuation,
        this_continuation_token=this_family_continuation,
        prev_continuation_token=prev_family_continuation,
        query_time_ms=None,
        total_time_ms=None,
    )


class VespaErrorDetails:
    """Wrapper for VespaError that parses the arguments"""

    def __init__(self, e: VespaError) -> None:
        self.e = e
        self.code = None
        self.summary = None
        self.message = None
        self.parse_args(self.e)

    def parse_args(self, e: VespaError) -> None:
        """
        Gets the details of the first error

        Args:
            e (VespaError): An error from the vespa python sdk
        """
        for arg in e.args:
            for error in arg:
                self.code = error.get("code")
                self.summary = error.get("summary")
                self.message = error.get("message")
                break

    @property
    def is_invalid_query_parameter(self) -> bool:
        """
        Checks if an error is coming from vespa on query parameters, see:

        https://github.com/vespa-engine/vespa/blob/0c55dc92a3bf889c67fac1ca855e6e33e1994904/
        container-core/src/main/java/com/yahoo/container/protect/Error.java
        """
        INVALID_QUERY_PARAMETER = 4
        return self.code == INVALID_QUERY_PARAMETER



from datetime import datetime
from enum import Enum
from typing import Any, List, Mapping, Optional, Sequence, Union

from pydantic import BaseModel, field_validator

Json = dict[str, Any]

CONTENT_TYPE_HTML = "text/html"
CONTENT_TYPE_DOCX = (
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
)
CONTENT_TYPE_PDF = "application/pdf"


class BackendDocument(BaseModel):
    """
    A representation of all information expected to be provided for a document.

    This class comprises direct information describing a document, along
    with all metadata values that should be associated with that document.
    """

    name: str
    document_title: Optional[str] = None
    description: str
    import_id: str
    slug: str
    family_import_id: str
    family_slug: str
    publication_ts: datetime
    date: Optional[str] = None  # Deprecated
    source_url: Optional[str] = None
    download_url: Optional[str] = None
    corpus_import_id: Optional[str] = None
    corpus_type_name: Optional[str] = None
    collection_title: Optional[str] = None
    collection_summary: Optional[str] = None
    type: str
    source: str
    category: str
    geography: str
    geographies: Optional[list[str]] = None
    languages: Sequence[str]

    metadata: Json

    @field_validator("type", mode="before")
    @classmethod
    def none_to_empty_string(cls, value):
        """If the value is None, will convert to an empty string"""
        return "" if value is None else value

    def to_json(self) -> Mapping[str, Any]:
        """Provide a serialisable version of the model"""

        json_dict = self.model_dump()
        json_dict["publication_ts"] = (
            self.publication_ts.isoformat() if self.publication_ts is not None else None
        )
        return json_dict


class InputData(BaseModel):
    """Expected input data containing RDS state."""

    documents: Mapping[str, BackendDocument]


class UpdateTypes(str, Enum):
    """
    UpdateTypes that are recognised and have resulting actions in the pipeline.

    A mapping of the update type to the action can be found in the ingest repo:
    https://github.com/climatepolicyradar/navigator-data-ingest/blob/main/src/
    navigator_data_ingest/base/updated_document_actions.py#L490

    Attributes:
        NAME (str): Represents the name of the document, causes embeddings generation to
            be re-triggered for a document.
        DESCRIPTION (str): Represents the description of the document, causes embeddings
            generation to be re-triggered for a document.
        SLUG (str): Represents the slug (a URL-friendly version of the name) of the
            document, triggers an update of the field in the relating s3 objects such
            that the new data is reflected in vespa.
        SOURCE_URL (str): Represents the source URL of the document and triggers full
            reprocessing and download from source of the document.
        METADATA (str): Represents the metadata associated with the document and
            indicates that the metadata of the objects in s3 relating to the document
            should be updated.
        REPARSE (str): Indicates that the document should be reparsed, including full
            reprocessing but not redownload from source.
        REPROCESS (str): Indicates that the document should be reprocessed, including
            redownload from source and reparse.
    """

    NAME = "name"
    DESCRIPTION = "description"
    SLUG = "slug"
    SOURCE_URL = "source_url"
    METADATA = "metadata"
    REPARSE = "reparse"
    REPROCESS = "reprocess"


class Update(BaseModel):
    """Results of comparing db state data against the s3 data to identify updates."""

    s3_value: Optional[Union[str, datetime, dict]] = None
    db_value: Optional[Union[str, datetime, dict]] = None
    type: UpdateTypes


class PipelineUpdates(BaseModel):
    """
    Expected input data containing document updates and new documents.

    This is utilized by the ingest stage of the pipeline.
    """

    new_documents: List[BackendDocument]
    updated_documents: dict[str, List[Update]]


class ExecutionData(BaseModel):
    """Data unique to a step functions execution that is required at later stages."""

    input_dir_path: str


class DocUpdateConfig(BaseModel):
    """
    Config for updates not defined as part of IdentifyUpdates.

    reprocess_updates: list of document ids to reprocess.
    """

    reprocess_updates: list[str]



import boto3


def get_aws_ssm_param(param_name: str, region_name: str = "eu-west-1") -> str:
    """Retrieve a parameter from AWS SSM"""
    ssm = boto3.client("ssm", region_name=region_name)
    response = ssm.get_parameter(Name=param_name, WithDecryption=True)
    return response["Parameter"]["Value"]



import csv
import re
from pathlib import Path
from typing import Any, Union


def is_sensitive_query(text: str, sensitive_terms: set) -> bool:
    """
    Scans text to determine if the query should be considered sensitive

    It does this by evaluating it against a set of predefined sensitive terms.
    These, as well as the specific logic are reproduced from previous work, which
    stated that "If the query contains any sensitive terms, and the length of the
    shortest sensitive term is >=50% of the length of the query by number of words..."
    then it is considered sensitive. Further details on the original can be found here:
    https://github.com/climatepolicyradar/navigator/pull/815

    This updated version builds on the above to avoid a loophole where a string of
    sensitive terms can be incorrectly flagged as not being sensitive. This happens
    because it is the shortest sensitive term that is compared to the rest of the
    query, and the rest of the query at that point can contain other sensitive terms.

    """
    sensitive_terms_in_query = [
        term for term in sensitive_terms if re.findall(term, text.lower())
    ]

    if sensitive_terms_in_query:
        terms = [term.pattern.strip("\\b") for term in sensitive_terms_in_query]
        shortest_sensitive_term = min(terms, key=len)
        shortest_sensitive_word_count = len(shortest_sensitive_term.split(" "))
        remaining_sensitive_word_count = sum(
            [len(term.split()) for term in terms if term != shortest_sensitive_term]
        )

        query_word_count = len(text.split())
        remaining_query_word_count = query_word_count - remaining_sensitive_word_count

        if remaining_query_word_count <= 0:
            return True

        proportion_sensitive = (
            shortest_sensitive_word_count / remaining_query_word_count
        )
        if proportion_sensitive >= 0.5:
            return True

    return False


def load_sensitive_query_terms() -> set[re.Pattern]:
    """
    Return sensitive query terms from the first column of a TSV file.

    Outputs are lowercased for case-insensitive matching.

    :return [set[str]]: sensitive query terms
    """
    tsv_path = Path(__file__).parent / "resources" / "sensitive_query_terms.tsv"
    with open(tsv_path, "r") as tsv_file:
        reader = csv.DictReader(tsv_file, delimiter="\t")
        sensitive_terms = []
        for row in reader:
            keyword = row["keyword"].lower().strip()
            keyword_regex = re.compile(r"\b" + re.escape(keyword) + r"\b")
            sensitive_terms.append(keyword_regex)
    return set(sensitive_terms)


def dig(obj: Union[list, dict], *fields: Any, default: Any = None) -> Any:
    """
    An interface for retrieving data from complicated objects

    Behaviour is to return the default if the path is invalid thereby avoiding errors
    Example: `dig(nested_dict, "parent", "child", "child_items", 1)`
    """
    for field in fields:
        if isinstance(obj, list):
            if isinstance(field, int) and len(obj) > field:
                obj = obj[field]
            else:
                return default
        elif isinstance(obj, dict):
            obj = obj.get(field, default)
        elif not obj:
            return default
    return obj


def unflatten_json(data: dict) -> dict:
    """
    Unflatten a dictionary with keys that are dot-separated strings.

    I.e. metadata.data respresents {"metadata": {"data": {}}}
    """
    unflattened = {}
    for key, value in data.items():
        parts = key.split(".")
        current = unflattened
        for part in parts[:-1]:
            current = current.setdefault(part, {})
        current[parts[-1]] = value
    return unflattened


def remove_key_if_all_nested_vals_none(data: dict, key: str) -> dict:
    """
    Remove the value for a given key if it's a dict with all None values.

    E.g. {"key": {"a": None, "b": None}} -> {}
    """
    if key not in data:
        return data
    if isinstance(data[key], dict):
        if all(value is None for value in data[key].values()):
            data.pop(key)
    return data
